import {
  Connection,
  PublicKey,
  Keypair,
  Transaction,
  VersionedTransaction,
  TransactionMessage,
  ComputeBudgetProgram,
  LAMPORTS_PER_SOL,
  SystemProgram,
} from "@solana/web3.js";
import {
  Liquidity,
  LiquidityPoolKeys,
  jsonInfo2PoolKeys,
  LiquidityPoolJsonInfo,
  TokenAccount,
  Token,
  TokenAmount,
  TOKEN_PROGRAM_ID,
  Percent,
  SPL_ACCOUNT_LAYOUT,
} from "@raydium-io/raydium-sdk";
import { Wallet } from "@coral-xyz/anchor";
import fetch from "node-fetch";
import { getSimulationComputeUnits } from "@solana-developers/helpers";
import { MICRO_LAMPORTS_PER_LAMPORTS, MINIMUM_TRANSACTION_FEE_IN_SOL, convertNumber } from "../utils";

export const loadPoolKeys = async (
  liquidityFile: string
): Promise<LiquidityPoolJsonInfo[] | undefined> => {
  const liquidityJsonResp = await fetch(liquidityFile);
  if (!liquidityJsonResp.ok) return;
  const liquidityJson = (await liquidityJsonResp.json()) as {
    official: any;
    unOfficial: any;
  };
  const allPoolKeysJson = [
    ...(liquidityJson?.official ?? []),
    ...(liquidityJson?.unOfficial ?? []),
  ];
  return allPoolKeysJson;
};

const SOL = "So11111111111111111111111111111111111111112";

/**
 * Class representing a Raydium Swap operation.
 */
class RaydiumSwap {
  allPoolKeysJson: LiquidityPoolJsonInfo[];
  connection: Connection;
  wallet: Wallet;

  /**
   * Create a RaydiumSwap instance.
   * @param {string} RPC_URL - The RPC URL for connecting to the Solana blockchain.
   * @param {string} WALLET_PRIVATE_KEY - The private key of the wallet in base58 format.
   */
  constructor(
    rpc: string,
    privateKey: Keypair,
    allPoolKeysJson: LiquidityPoolJsonInfo[]
  ) {
    this.connection = new Connection(rpc, { commitment: "confirmed" });
    this.wallet = new Wallet(privateKey);
    this.allPoolKeysJson = allPoolKeysJson;
  }

  /**
   * Loads all the pool keys available from a JSON configuration file.
   * @async
   * @returns {Promise<void>}
   */
  async loadPoolKeys(liquidityFile: string) {
    if (this.allPoolKeysJson.length > 0) return;

    const liquidityJsonResp = await fetch(liquidityFile);
    if (!liquidityJsonResp.ok) return;
    const liquidityJson = (await liquidityJsonResp.json()) as {
      official: any;
      unOfficial: any;
    };
    const allPoolKeysJson = [
      ...(liquidityJson?.official ?? []),
      ...(liquidityJson?.unOfficial ?? []),
    ];

    this.allPoolKeysJson = allPoolKeysJson;
  }

  /**
   * Finds pool information for the given token pair.
   * @param {string} mintA - The mint address of the first token.
   * @param {string} mintB - The mint address of the second token.
   * @returns {LiquidityPoolKeys | null} The liquidity pool keys if found, otherwise null.
   */
  findPoolInfoForTokens(mintA: string, mintB: string) {
    const poolData = this.allPoolKeysJson.find(
      (i) =>
        (i.baseMint === mintA && i.quoteMint === mintB) ||
        (i.baseMint === mintB && i.quoteMint === mintA)
    );

    if (!poolData) return null;

    return jsonInfo2PoolKeys(poolData) as LiquidityPoolKeys;
  }

  /**
   * Retrieves token accounts owned by the wallet.
   * @async
   * @returns {Promise<TokenAccount[]>} An array of token accounts.
   */
  async getOwnerTokenAccounts() {
    const walletTokenAccount = await this.connection.getTokenAccountsByOwner(
      this.wallet.publicKey,
      {
        programId: TOKEN_PROGRAM_ID,
      }
    );

    return walletTokenAccount.value.map((i) => ({
      pubkey: i.pubkey,
      programId: i.account.owner,
      accountInfo: SPL_ACCOUNT_LAYOUT.decode(i.account.data),
    }));
  }

  /**
   * Builds a swap transaction.
   * @async
   * @param {string} toToken - The mint address of the token to receive.
   * @param {number} amount - The amount of the token to swap.
   * @param {LiquidityPoolKeys} poolKeys - The liquidity pool keys.
   * @param {string} slippage - The slippage percentage.
   * @param {string | null} priceImpact - The price impact percentage.
   * @param {string | null} fee - The fee percentage.
   * @param {number} [maxLamports=100000] - The maximum lamports to use for transaction fees.
   * @param {boolean} [useVersionedTransaction=true] - Whether to use a versioned transaction.
   * @param {'in' | 'out'} [fixedSide='in'] - The fixed side of the swap ('in' or 'out').
   * @returns {Promise<Transaction | VersionedTransaction>} The constructed swap transaction.
   */
  async getSwapTransaction(
    toToken: string,
    amount: number,
    poolKeys: LiquidityPoolKeys,
    slippage: string = "5",
    priceImpact: string | null = null,
    fee: string | null = null,
    feePercent: string,
    feeRecipient: PublicKey,
    maxLamports: number = 100000,
    // useVersionedTransaction = false,
    fixedSide: "in" | "out" = "in",
    // ): Promise<{ transaction: Transaction | VersionedTransaction, fee: number }> {
  ): Promise<{ transaction: Transaction, amountOut: number, fee: number }> {
    const directionIn = poolKeys.quoteMint.toString() == toToken;
    const { amountOut, minAmountOut, amountIn } = await this.calcAmountOut(
      poolKeys,
      amount,
      directionIn,
      slippage,
      priceImpact,
    );
    const userTokenAccounts = await this.getOwnerTokenAccounts();
    const swapTransaction = await Liquidity.makeSwapInstructionSimple({
      connection: this.connection,
      // makeTxVersion: useVersionedTransaction ? 0 : 1,
      makeTxVersion: 1,
      poolKeys: {
        ...poolKeys,
      },
      userKeys: {
        tokenAccounts: userTokenAccounts,
        owner: this.wallet.publicKey,
      },
      amountIn: amountIn,
      amountOut: minAmountOut,
      fixedSide: fixedSide,
      config: {
        bypassAssociatedCheck: false,
      },
      computeBudgetConfig: {
        microLamports: maxLamports,
      },
    });

    const instructions =
      swapTransaction.innerTransactions[0].instructions.filter(Boolean);

    let feeAmountInLamports = 0;
    // Check if fee is enabled
    if (feePercent && parseFloat(feePercent) > 0) {
      let feeAmount = parseFloat("0");
      if (toToken !== SOL) {
        feeAmount = parseFloat(amount.toString()) * parseFloat(feePercent);
        feeAmountInLamports = Math.round(feeAmount * LAMPORTS_PER_SOL);
      } else {
        feeAmount = parseFloat(amountOut.toExact()) * parseFloat(feePercent);
        feeAmountInLamports = Math.round(feeAmount * LAMPORTS_PER_SOL);
      }

      if (feeAmountInLamports > 0) {
        instructions.push(
          SystemProgram.transfer({
            fromPubkey: this.wallet.publicKey,
            toPubkey: feeRecipient,
            lamports: feeAmountInLamports,
          })
        );
      }
    }

    const units = await getSimulationComputeUnits(this.connection, instructions, this.wallet.payer.publicKey, []);
    // Remove fake setComputeUnitPrice
    instructions.shift();
    let swapFee = MINIMUM_TRANSACTION_FEE_IN_SOL;
    let computeUnitPrice = 0;

    if (fee && units) {
      swapFee = convertNumber(fee);
      if (swapFee >= MINIMUM_TRANSACTION_FEE_IN_SOL) {
        const diffFee = swapFee - MINIMUM_TRANSACTION_FEE_IN_SOL;
        computeUnitPrice = Math.round((diffFee * LAMPORTS_PER_SOL / units) * MICRO_LAMPORTS_PER_LAMPORTS);
      } else {
        throw Error("Fee must be greater than 5000 lamports (0.000005 SOL)");
      }
    }

    if (computeUnitPrice > 0 && units) {
      instructions.unshift(ComputeBudgetProgram.setComputeUnitPrice({
        microLamports: computeUnitPrice
      }))

      instructions.unshift(ComputeBudgetProgram.setComputeUnitLimit({
        units
      }))
    }

    const recentBlockhashForSwap = await this.connection.getLatestBlockhash();
    // if (useVersionedTransaction) {
    //   const versionedTransaction = new VersionedTransaction(
    //     new TransactionMessage({
    //       payerKey: this.wallet.publicKey,
    //       recentBlockhash: recentBlockhashForSwap.blockhash,
    //       instructions: instructions,
    //     }).compileToV0Message()
    //   );

    //   versionedTransaction.sign([this.wallet.payer]);

    //   return { transaction: versionedTransaction, fee: feeAmount };
    // }

    const legacyTransaction = new Transaction({
      blockhash: recentBlockhashForSwap.blockhash,
      lastValidBlockHeight: recentBlockhashForSwap.lastValidBlockHeight,
      feePayer: this.wallet.publicKey,
    });

    legacyTransaction.add(...instructions);

    return { transaction: legacyTransaction, amountOut: parseFloat(amountOut.toExact()), fee: feeAmountInLamports / LAMPORTS_PER_SOL };
  }

  /**
   * Sends a legacy transaction.
   * @async
   * @param {Transaction} tx - The transaction to send.
   * @returns {Promise<string>} The transaction ID.
   */
  async sendLegacyTransaction(tx: Transaction, maxRetries?: number) {
    const txid = await this.connection.sendTransaction(
      tx,
      [this.wallet.payer],
      {
        skipPreflight: true,
        maxRetries: maxRetries,
      }
    );

    return txid;
  }

  /**
   * Sends a versioned transaction.
   * @async
   * @param {VersionedTransaction} tx - The versioned transaction to send.
   * @returns {Promise<string>} The transaction ID.
   */
  async sendVersionedTransaction(
    tx: VersionedTransaction,
    maxRetries?: number
  ) {
    const txid = await this.connection.sendTransaction(tx, {
      skipPreflight: true,
      maxRetries: maxRetries,
    });

    return txid;
  }

  /**
   * Simulates a versioned transaction.
   * @async
   * @param {VersionedTransaction} tx - The versioned transaction to simulate.
   * @returns {Promise<any>} The simulation result.
   */
  async simulateLegacyTransaction(tx: Transaction) {
    const txid = await this.connection.simulateTransaction(tx, [
      this.wallet.payer,
    ]);

    return txid;
  }

  /**
   * Simulates a versioned transaction.
   * @async
   * @param {VersionedTransaction} tx - The versioned transaction to simulate.
   * @returns {Promise<any>} The simulation result.
   */
  async simulateVersionedTransaction(tx: VersionedTransaction) {
    const txid = await this.connection.simulateTransaction(tx);

    return txid;
  }

  /**
   * Gets a token account by owner and mint address.
   * @param {PublicKey} mint - The mint address of the token.
   * @returns {TokenAccount} The token account.
   */
  getTokenAccountByOwnerAndMint(mint: PublicKey) {
    return {
      programId: TOKEN_PROGRAM_ID,
      pubkey: PublicKey.default,
      accountInfo: {
        mint: mint,
        amount: 0,
      },
    } as unknown as TokenAccount;
  }

  /**
   * Calculates the amount out for a swap.
   * @async
   * @param {LiquidityPoolKeys} poolKeys - The liquidity pool keys.
   * @param {number} rawAmountIn - The raw amount of the input token.
   * @param {boolean} swapInDirection - The direction of the swap (true for in, false for out).
   * @returns {Promise<Object>} The swap calculation result.
   */
  async calcAmountOut(
    poolKeys: LiquidityPoolKeys,
    rawAmountIn: number,
    swapInDirection: boolean,
    expectSlippage: string = "5", // 5%
    expectPriceImpact: string | null = null
  ) {
    const poolInfo = await Liquidity.fetchInfo({
      connection: this.connection,
      poolKeys,
    });

    let currencyInMint = poolKeys.baseMint;
    let currencyInDecimals = poolInfo.baseDecimals;
    let currencyOutMint = poolKeys.quoteMint;
    let currencyOutDecimals = poolInfo.quoteDecimals;

    if (!swapInDirection) {
      currencyInMint = poolKeys.quoteMint;
      currencyInDecimals = poolInfo.quoteDecimals;
      currencyOutMint = poolKeys.baseMint;
      currencyOutDecimals = poolInfo.baseDecimals;
    }

    const currencyIn = new Token(
      TOKEN_PROGRAM_ID,
      currencyInMint,
      currencyInDecimals
    );
    const amountIn = new TokenAmount(currencyIn, rawAmountIn, false);
    const currencyOut = new Token(
      TOKEN_PROGRAM_ID,
      currencyOutMint,
      currencyOutDecimals
    );
    const slippage = new Percent(Number(expectSlippage), 100);

    const {
      amountOut,
      minAmountOut,
      currentPrice,
      executionPrice,
      priceImpact,
      fee,
    } = Liquidity.computeAmountOut({
      poolKeys,
      poolInfo,
      amountIn,
      currencyOut,
      slippage,
    });
    if (expectPriceImpact) {
      const parseExpectPriceImpact = parseFloat(expectPriceImpact);
      const parsePriceImpact = parseFloat(priceImpact.toSignificant());

      if (parsePriceImpact > parseExpectPriceImpact) {
        throw new Error("Price impact is too high");
      }
    }

    return {
      amountIn,
      amountOut,
      minAmountOut,
      currentPrice,
      executionPrice,
      priceImpact,
      fee,
    };
  }
}

export default RaydiumSwap;
